/*
 * Simple Ogg Vorbis decoder, (c) 2009 WWRN Vincent B Washington
 *
 * Based in part on the original sample ogg vorbis decoder - structure is
 * designed for reuse.
 *
 * Incorporates code (c) 2004 BBC Michael Sparks
 * Incorporates code (c) 1994-2004 XIPHOPHORUS Company http://www.xiph.org/
 *
 * Important limitations include:
 *   * Has a minimum buffer size of 58 bytes (important detail)
 *  <*> This is an inherent system design
 *   * Does not handle chained bitstreams.
 *     (Multiple vorbis streams concatenated in a file)
 *  <*> Uses ogg_page_eos function to report end of stream
 *   * Uses EOF to indicate end of stream
 *  <*> Uses EOF to indicate end of chain
 *
 * Seems to decode most ogg vorbis file pretty well though
 *
 */

#include "oggvorbis_StateMachine.h"

Machine *StateMachine;		/* Mathematical equvalent For STATE MACHINE f(x) = sm; y = RTN val */
goNogo DigitalWordRegister;
oggVorbisContext * oggVorbisBlackBox;
source_buffer * sourceBuffer;
decode_buffer * decodeBuffer;
FILE *DigitalWordRegisterSchdule;

/****************************************************************************
 * Code for handling ogg vorbis contexts
 */

oggVorbisContext * newOggVorbisContext(void) {
   oggVorbisContext * result;
   result = (oggVorbisContext *)malloc(sizeof(oggVorbisContext));
   if (result) {
      memset(result, 0, sizeof(oggVorbisContext));
      ogg_sync_init(&(result->oggSync)); /* Now we can read pages */
      DigitalWordRegister = (DigitalBit6 | (DigitalWordRegister & ResetITLA));
      result->headerPacketsRead = 0;
      result->streamInitialised = 0;
      result->convsize = 0;
      result->buffer = NULL;
      result->bufferlen = 0;
      result->warnClipping = FALSE;
      result->decodeState = TRUE;
      result->e_o_s = FALSE;
      /* result->oggBitstreamPage; */ /* No initialisation */
      return result;
   }
   return NULL;
}

/*************************************************************************
 * Code for dealing with source buffer data
 */

source_buffer * newSourceBuffer(FILE * fh, int buffersize) {
   source_buffer * result;
   result = (source_buffer *)malloc(sizeof(source_buffer));
   if (result) {
      memset(result, 0, sizeof(source_buffer));
      result->fh = fh;
      result->bytes = 0;
      if (buffersize) {
         result->buffer = (char *)malloc(sizeof(char)*buffersize);
      } else {
         result->buffer = NULL;
      }
      result->buffersize = buffersize;
      return result;
   }
   return NULL;
}

/****************************************************************************
 * Code for handling decode buffer results
 */

goNogo newDecodeBuffer(void) {
   decodeBuffer = (decode_buffer *)malloc(sizeof(decode_buffer));
   if (decodeBuffer) {
      memset(decodeBuffer, 0, sizeof(decode_buffer));
      DigitalWordRegister = (DigitalBit1 | (DigitalWordRegister & ResetETLA));
      return 0;
   }
   return 1;
}

goNogo sendBytesForDecode(void){
   if ((oggVorbisBlackBox->enoughDataPages=ogg_sync_pageout(&(oggVorbisBlackBox->oggSync),&(oggVorbisBlackBox->oggBitstreamPage))) !=1)
   {
       char * buffer;
       buffer=ogg_sync_buffer(&(oggVorbisBlackBox->oggSync),sourceBuffer->buffersize);
       memcpy(buffer,sourceBuffer->buffer, (sizeof(char)*sourceBuffer->bytes));
       ogg_sync_wrote(&(oggVorbisBlackBox->oggSync),sourceBuffer->bytes);
   }
}

goNogo OggVorbisVerificationTest(void) {
   int result;
   if (oggVorbisBlackBox->enoughDataPages<1) {
       oggVorbisBlackBox->enoughDataPages=ogg_sync_pageout(&(oggVorbisBlackBox->oggSync),&(oggVorbisBlackBox->oggBitstreamPage));
   }
   if(ogg_page_bos(&oggVorbisBlackBox->oggBitstreamPage)) fprintf(stderr, "Start of Stream\n");
   if (! oggVorbisBlackBox->streamInitialised) {
      ogg_stream_init(&(oggVorbisBlackBox->oggStream), ogg_page_serialno(&(oggVorbisBlackBox->oggBitstreamPage)));
      oggVorbisBlackBox->streamInitialised = TRUE;
   }

   vorbis_info_init(&(oggVorbisBlackBox->vorbisInfo));
   vorbis_comment_init(&(oggVorbisBlackBox->vorbisComment));
   if ((result = readPacket())<1) goto end;
   result=vorbis_synthesis_headerin(&(oggVorbisBlackBox->vorbisInfo),&(oggVorbisBlackBox->vorbisComment),&(oggVorbisBlackBox->oggPacket));
   if (result<0) goto end;
   else result=1;

   end:
      if (result>0) {
          DigitalWordRegister = (DigitalBit8 | (DigitalWordRegister & ResetITLA));
      }
      if (result==0) {
          DigitalWordRegister = (DigitalBit7 | (DigitalWordRegister & ResetITLA));
          DigitalWordRegister = (DigitalBit2 | (DigitalWordRegister & ResetETLA));
      }
      if (result <0)  DigitalWordRegister = (DigitalBit11 | (DigitalWordRegister & (ResetITLA & ResetETLA)));
   return result; 
}

goNogo readPacket(void){
   int packetResult=0, pageResult;
   
   if ((pageResult=ogg_stream_pagein(&(oggVorbisBlackBox->oggStream),&(oggVorbisBlackBox->oggBitstreamPage)))<0) {return CORRUPTPAGE;}
   if ((packetResult=ogg_stream_packetout(&(oggVorbisBlackBox->oggStream),&(oggVorbisBlackBox->oggPacket)))<0) {return CORRUPTPACKET;}
   else {packetResult=PACKETREAD;}

   return packetResult;
}

goNogo CodeBookAndPreMetaData(void) {
   int result;
    
   while (oggVorbisBlackBox->headerPacketsRead<2) {
      if ((result=ogg_sync_pageout(&(oggVorbisBlackBox->oggSync),&(oggVorbisBlackBox->oggBitstreamPage)))>0) {
          ogg_stream_pagein(&(oggVorbisBlackBox->oggStream),&(oggVorbisBlackBox->oggBitstreamPage));
          while((result=ogg_stream_packetout(&(oggVorbisBlackBox->oggStream),&(oggVorbisBlackBox->oggPacket)))>0) {
              vorbis_synthesis_headerin(&(oggVorbisBlackBox->vorbisInfo),&(oggVorbisBlackBox->vorbisComment),&(oggVorbisBlackBox->oggPacket));
               if ((oggVorbisBlackBox->headerPacketsRead += 1)>1) break;
           }
           if (result<0) goto end;
      }
      else {result=0; goto end;}
   }
   oggVorbisBlackBox->headerPacketsRead=0;
   result=COMMENTANDCODEBOOKHEADERSREAD;  
  {
      char **cptr=oggVorbisBlackBox->vorbisComment.user_comments;
      while(*cptr){

	fprintf(stderr,"%s\n",*cptr);
	++cptr;
      }
      fprintf(stderr,"\nBitstream is %d channel, %ldHz\n",oggVorbisBlackBox->vorbisInfo.channels,oggVorbisBlackBox->vorbisInfo.rate);
      fprintf(stderr,"Encoded by: %s\n\n",oggVorbisBlackBox->vorbisComment.vendor);
  }
  oggVorbisBlackBox->convsize=BUFSIZE/oggVorbisBlackBox->vorbisInfo.channels;
  vorbis_synthesis_init(&(oggVorbisBlackBox->vorbisDSP),&(oggVorbisBlackBox->vorbisInfo));
  vorbis_block_init(&(oggVorbisBlackBox->vorbisDSP),&(oggVorbisBlackBox->vorbisWorkingBlock));
  
  end:
      if (result==1) {
        DigitalWordRegister = (DigitalBit9 | (DigitalWordRegister & ResetITLA));
       //	DigitalWordRegister = (DigitalBit2 | (DigitalWordRegister & ResetETLA));
      }
      if (result==0) {
        DigitalWordRegister = (DigitalBit2 | (DigitalWordRegister & ResetETLA));
      }
      if (result <0)  DigitalWordRegister = (DigitalBit11 | (DigitalWordRegister & (ResetITLA & ResetETLA)));
}

goNogo PCMfloatsToIntPCM(int samples, float **pcm, int bout) {
    int j;
    int i;
    int clipflag=0;
    int channels = oggVorbisBlackBox->vorbisInfo.channels;
    for(i=0;i<channels;i++){
        ogg_int16_t *ptr=(oggVorbisBlackBox->convbuffer)+i;
        float  *mono=pcm[i];
	//float  *mono=*pcm;
        for(j=0;j<bout;j++){
            #if 0
                int val=mono[j]*32767.f;
            #else /* optional dither */
                int val=mono[j]*32767.f+drand48()-0.5f;
            #endif
            /* might as well guard against clipping */
            if(val>32767){
                val=32767;
                clipflag=1;
            }
            if(val<-32768){
                val=-32768;
                clipflag=1;
            }
            *ptr=val;
            ptr+=channels;
        }
    }
    return clipflag;
}

goNogo AudioPacketDecodeSynthesis(void) {
//fprintf(stderr, "DecodeDataStream\n" );
   int vs_result;
   int result, oggOne=0, oggTwo=0;
   int convsize = oggVorbisBlackBox->convsize;
   float **pcm;
   int samples;
   goNogo havedata=FALSE;

while (oggVorbisBlackBox->decodeState) {
   
   if (oggVorbisBlackBox->decodeState&1) {
     while((result=ogg_sync_pageout(&oggVorbisBlackBox->oggSync,&oggVorbisBlackBox->oggBitstreamPage))<1){
       if(result<1) {result=DigitalBit2; goto catchall; }  // need more data 
     } 
     oggVorbisBlackBox->decodeState=2;
     ogg_stream_pagein(&oggVorbisBlackBox->oggStream,&oggVorbisBlackBox->oggBitstreamPage); /* can safely ignore errors at this point */
   }
     
   if (oggVorbisBlackBox->decodeState&2) {
     while((result=ogg_stream_packetout(&oggVorbisBlackBox->oggStream,&oggVorbisBlackBox->oggPacket))<1){
       if(result==0) { oggVorbisBlackBox->decodeState=1; result=DigitalBit2; goto catchall; }
       if(result<0) continue;
     } 
     clearBuffer();
     if ((vorbis_synthesis(&(oggVorbisBlackBox->vorbisWorkingBlock),
                                 &(oggVorbisBlackBox->oggPacket)))==0)
       vorbis_synthesis_blockin(&(oggVorbisBlackBox->vorbisDSP),&(oggVorbisBlackBox->vorbisWorkingBlock));

     havedata=FALSE;

     while((samples=vorbis_synthesis_pcmout(&(oggVorbisBlackBox->vorbisDSP),&pcm))>0) {
       int clipflag=0;
       int bout=(samples<convsize?samples:convsize);
       havedata=TRUE; 
       clipflag = PCMfloatsToIntPCM(samples, pcm, bout);
       if(oggVorbisBlackBox->warnClipping && clipflag)
            fprintf(stderr,"Clipping in frame %ld\n",(long)(oggVorbisBlackBox->vorbisDSP.sequence));

       bufferResults(bout);
          vorbis_synthesis_read(&(oggVorbisBlackBox->vorbisDSP),bout);
     }
     oggVorbisBlackBox->decodeState=2;
     if (havedata) { result=DigitalBit3; goto catchall; } // If we entered the loop it means we have data
   }
}
   catchall:
        if (result==DigitalBit2)
            if (oggVorbisBlackBox->e_o_s) {
               vorbis_block_clear(&oggVorbisBlackBox->vorbisWorkingBlock);
               vorbis_dsp_clear(&oggVorbisBlackBox->vorbisDSP);
               ogg_stream_clear(&oggVorbisBlackBox->oggStream);
               vorbis_comment_clear(&oggVorbisBlackBox->vorbisComment);
               vorbis_info_clear(&oggVorbisBlackBox->vorbisInfo);  /* must be called last */
               oggVorbisBlackBox->streamInitialised = FALSE;
               oggVorbisBlackBox->decodeState=1;
               DigitalWordRegister = (DigitalBit1 | (DigitalWordRegister & ResetETLA));
               DigitalWordRegister = (DigitalBit6 | (DigitalWordRegister & ResetITLA));
               fprintf(stderr, "End of Stream\n");
            }
            else DigitalWordRegister = (DigitalBit2 | (DigitalWordRegister & ResetETLA));
        if (result == DigitalBit3) {
            if (oggVorbisBlackBox->bufferlen>0) {
               /* We pass back aliases to the internal passback buffers */
               decodeBuffer->buffer = oggVorbisBlackBox->buffer;
               decodeBuffer->len = oggVorbisBlackBox->bufferlen;
               DigitalWordRegister = (DigitalBit3 | (DigitalWordRegister & ResetETLA));
               DigitalWordRegister = (DigitalBit10 | (DigitalWordRegister & ResetITLA));
            }
        }
//fprintf(stderr, "oggVorbisBlackBox->decodeState Leaving[%d] result[0x%X]\n", oggVorbisBlackBox->decodeState, result);
}

void clearBuffer(void){
   if (oggVorbisBlackBox->buffer)
      free(oggVorbisBlackBox->buffer);
   oggVorbisBlackBox->buffer = NULL;
   oggVorbisBlackBox->bufferlen = 0;
}

void bufferResults(int bout){
   char * buffer;
   int len = 2*(oggVorbisBlackBox->vorbisInfo.channels)*bout;
   if (! oggVorbisBlackBox->buffer) {
      buffer = (char *) malloc(len);
      memcpy((void *) buffer, (void *) oggVorbisBlackBox->convbuffer, len);
      oggVorbisBlackBox->buffer = buffer;
      oggVorbisBlackBox->bufferlen = len;
   }
    else {
      buffer = (char *) malloc(oggVorbisBlackBox->bufferlen + len);
      memcpy((void *) buffer, (void *) oggVorbisBlackBox->buffer, oggVorbisBlackBox->bufferlen);

      free(oggVorbisBlackBox->buffer);        
      oggVorbisBlackBox->buffer = buffer ;    
      buffer += oggVorbisBlackBox->bufferlen; 
      oggVorbisBlackBox->bufferlen += len;    
      memcpy((void *) buffer, (void *) oggVorbisBlackBox->convbuffer, len);
   }
}

goNogo processAudio(void){

  if (newDecodeBuffer()) { DigitalWordRegister = (DIE | (DigitalWordRegister & ResetETLA));} 
  StateMachine->CallRoutine[LookUpTable(DigitalWordRegister & ITLADECODING)]();
   return 0;
}

void readData(void){
   sourceBuffer->bytes = -1;
   if (!sourceBuffer->buffer) {
      fprintf(stderr, "No buffer to read data into\n");
      return;
   }
   sourceBuffer->bytes = fread(sourceBuffer->buffer,1,sourceBuffer->buffersize, sourceBuffer->fh);
}

goNogo getDATA(void)
{
   readData();
   if (sourceBuffer->bytes>1)
        sendBytesForDecode();
   else {DigitalWordRegister = (DIE | (DigitalWordRegister & ResetETLA)); return 0;}

   DigitalWordRegister = (DigitalBit1 | (DigitalWordRegister & ResetETLA));
   return 0;
}
//FILE *DNULL=fopen("/dev/null", "w");
goNogo sendDATA(void)
{
int bytes;
   bytes=fwrite(decodeBuffer->buffer,sizeof(char),decodeBuffer->len,stdout);
   DigitalWordRegister = (DigitalBit1 | (DigitalWordRegister & ResetETLA));
   if(ogg_stream_eos(&oggVorbisBlackBox->oggStream)) {oggVorbisBlackBox->e_o_s=1;

   }
   if ((DigitalWordRegister & ITLADECODING) == DigitalBit10)
	DigitalWordRegister = (DigitalBit9 | (DigitalWordRegister & ResetITLA));
   return 0;
}

goNogo initCALL(void)
{
   DigitalWordRegister=CLEAR;
   if ((oggVorbisBlackBox = newOggVorbisContext()) != NULL)
   {
        if ((sourceBuffer = newSourceBuffer(stdin,BUFSIZE)) == NULL)
	{
		free(sourceBuffer);
		exit(0);
	}
	ogg_sync_init(&oggVorbisBlackBox->oggSync);
	DigitalWordRegister = (DigitalBit5 | (DigitalWordRegister & ResetETLA));
	return 0;
   }
   exit(0);
}

goNogo recycle(void){ if (decodeBuffer != NULL) { free(decodeBuffer); decodeBuffer = NULL; } DigitalWordRegister = (DigitalBit5 | (DigitalWordRegister & ResetETLA)); }

goNogo Commence(void)
{
   DigitalWordRegister = (DigitalBit2 | (DigitalWordRegister & ResetETLA));
   DigitalWordRegister = (DigitalBit7 | (DigitalWordRegister & ResetITLA));
return 0;
}

goNogo LookUpTable(int n)
{
   int rtn;	
   if (n==1)
	rtn=1;	
   else if (n==2)
	rtn=2;
   else if (n==4)
	rtn=3;
   else if (n==8)
	rtn=4;
   else if (n==16)
	rtn=5;
   else if (n==32)
	rtn=6;
   else if (n==64)
	rtn=7;
   else if (n==128)
	rtn=8;
   else if (n==256)
	rtn=9;
return(rtn);
}
