/*
 * Simple Ogg Vorbis decoder, (c) 2009 WWRN Vincent B Washington
 *
 * Based in part on the original sample ogg vorbis decoder - structure is
 * designed for reuse.
 *
 * Incorporates code (c) 2004 BBC Michael Sparks
 * Incorporates code (c) 1994-2004 XIPHOPHORUS Company http://www.xiph.org/
 *
 * Important limitations include:
 *   * Has a minimum buffer size of 58 bytes (important detail)
 *   * Does not handle chained bitstreams.
 *     (Multiple vorbis streams concatenated in a file)
 *   * Uses EOF to indicate end of stream
 *
 * Seems to decode most ogg vorbis file pretty well though
 *
 */
#ifndef _OGGVORBISSTATEMACHINE_H_
#define _OGGVORBISSTATEMACHINE_H_

#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
// FIXME: Currently the minimum buffer size is 58. (needs work!)
#define BUFSIZE 4096

#define FALSE 0x00000000
#define TRUE 0x00000001

// Logic goNogo (s)
#define PACKETREAD TRUE
#define PAGEREAD TRUE
#define CLEAR FALSE
#define COMMENTANDCODEBOOKHEADERSREAD TRUE

// Exterior Top Level Enablers
#define DigitalBit1 0x00000001		/* tryAgain */
#define DigitalBit2 0x00000002		/* sendSystemMoreDATA */
#define DigitalBit3 0x00000004		/* sendToOutputBuffer */
#define DigitalBit4 0x00000008		/* initStateCall */
#define DigitalBit5 0x0000010		/* processAudio */

// Interior Top Level Enablers
#define DigitalBit6 0x00000020		/* COMMENCE */
#define DigitalBit7 0x00000040		/* TESTFORFIRSTHEADER */
#define DigitalBit8 0x00000080		/* CODEBOOKANDPREMETADATA */
#define DigitalBit9 0x00000100		/* PROCESSOUTPUTDATA */
#define DigitalBit10 0x00000200		/* READYTOOUTPUTDATA */

// Decode ERRORS
#define DigitalBit11 0x00000400		/* FAILSTATE */
#define CORRUPTPAGE -2
#define CORRUPTPACKET -3
#define NOTOGGVORBIS -1
#define DIE FALSE

 // Enablers and Resets
#define ETLADECODING 0x0000001F
#define ResetETLA 0xFFFFFE0
#define ITLADECODING 0x000003E0
#define ResetITLA 0xFFFFE01F


typedef int goNogo;

typedef struct {
   int              decodeState;
   goNogo           enoughDataPackages;
   goNogo           enoughDataPages;
   ogg_sync_state   oggSync;              /* sync and verify incoming physical bitstream */
   ogg_page         oggBitstreamPage;     /* one Ogg bitstream page.  Vorbis packets are inside */
   ogg_stream_state oggStream;            /* take physical pages, weld into a logical stream of packets */
   ogg_packet       oggPacket;            /* one raw packet of data for decode */
   vorbis_info      vorbisInfo;           /* struct that stores all the static vorbis bitstream settings */
   vorbis_comment   vorbisComment;        /* struct that stores all the bitstream user comments */
   vorbis_dsp_state vorbisDSP;            /* central working state for the packet->PCM decoder */
   vorbis_block     vorbisWorkingBlock;   /* local working space for packet->PCM decode */
   ogg_int16_t      convbuffer[BUFSIZE];  /* take 8k out of the data segment, not the stack */
   long             b_o_s;
   long             e_o_s;
   goNogo streamInitialised;
   goNogo warnClipping;
   int headerPacketsRead;
   int convsize;
   char * buffer; // Intermediate decode buffer (for passback to client)
   int    bufferlen;
   //Machine *BlackBox;
} oggVorbisContext;

typedef struct {
   FILE * fh;
   char * buffer;
   int bytes;
   int buffersize;
} source_buffer;

typedef struct {
   char * buffer; // The actual storage space
   int len;
   //int status;
} decode_buffer;


typedef struct _Machine {
   goNogo (*CallRoutine[17])(void);
} Machine;

extern Machine *StateMachine;		/* Mathematical equvalent For STATE MACHINE f(x)= sm; y =RTN val */
extern goNogo DigitalWordRegister;
extern oggVorbisContext * oggVorbisBlackBox;
extern source_buffer * sourceBuffer;
extern decode_buffer * decodeBuffer;


extern FILE *DigitalWordRegisterSchdule;



/* Prototypes */
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* WRAPPER FUNCTION */ extern goNogo DigitalWordProcessor(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ extern goNogo initCALL(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ extern goNogo recycle(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* WRAPPER FUNCTION */ extern goNogo getDATA(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ extern goNogo sendDATA(void);

/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* WRAPPER FUNCTION */ extern goNogo Commence(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo OggVorbisVerificationTest(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo AudioPacketDecodeSynthesis(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo CodeBookAndPreMetaData(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ //extern goNogo MiniStateMachineDecoder(void);

/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern oggVorbisContext * newOggVorbisContext(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern source_buffer * newSourceBuffer(FILE * fh, int buffersize);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo newDecodeBuffer(void) ;
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern void readData(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo sendBytesForDecode(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo processAudio(void);

/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo outputPackage(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo readPacket(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo PCMfloatsToIntPCM(int samples, float **pcm, int bout);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern void clearBuffer(void);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern void bufferResults(int bout);
/* oggvorbis_StateMachine.c */ /* DOCUMENTED */ /* BlackBox */ extern goNogo LookUpTable(int n);


#endif
